/**
 * @file SeedService.hh
 * @brief An art service to assist in the distribution of guaranteed unique seeds to all engines within an art job.
 * @author Rob Kutschke (kutschke@fnal.gov)
 * @date   2013/03/14 19:54:49
 * @see SeedService_service.cc
 * 
 */

/******************************************************************************
 *** TODO QA
 *** 
 *** - the original code had copy constructor made private to prevent it from
 ***   being called (now it is deleted); is this a prescription for a service?
 *** 
 ******************************************************************************/

#ifndef SeedService_SeedService_hh
#define SeedService_SeedService_hh

// C/C++ standard libraries
#include <string>
#include <map>
#include <iomanip>

// Some helper classes.
#include "artextensions/SeedService/ArtState.hh"
#include "artextensions/SeedService/SeedMaster.hh"

// From art and its tool chain.
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"

// Forward declarations
namespace art {
  class ActivityRegistry;
  class ModuleDescription;
  class Run;
  class SubRun;
}

/// artextensions repository namespace
namespace artext {
  
  /**
   * @brief An art service to assist in the distribution of guaranteed unique seeds to all engines within an art job.
   * @see SeedMaster
   *
   * The SeedService acts as an interface between art framework and the
   * SeedMaster class.
   * 
   * The documentation is mantained in the SeedMaster class.
   * The configuration of SeedService is exactly the same as SeedMaster's,
   * and in art it's read from `services.user.SeedService`.
   * In a art module, code requests a seed by making one of the following two
   * calls:
   *     
   *     art::ServiceHandle<SeedService>->getSeed();
   *     art::ServiceHandle<SeedService>->getSeed("instanceName");
   *     
   * This must happen in the constructor of the module.
   * The reason is that we don't want engines to be initialized in the middle
   * of a job.
   */
  class SeedService {
  public:
    
    using seed_t = art::RandomNumberGenerator::seed_t;
    
    SeedService(const fhicl::ParameterSet&, art::ActivityRegistry&);
    
    // Accept compiler written d'tor.  Not copyable or assignable.
    // This class is not copyable or assignable: these methods are not implemented.
    SeedService(SeedService const&) = delete;
    SeedService const& operator=(SeedService const&) = delete;
    SeedService(SeedService&&) = delete;
    SeedService const& operator=(SeedService&&) = delete;
    ~SeedService() = default;
    
    /// Returns the seed value for this module label (instance name).
    seed_t getSeed();
    seed_t getSeed(std::string instanceName);
    
    /// Prints known (EngineId,seed) pairs.
    template<class Stream> void print(Stream&) const;
    
    /// Prints to the framework Info logger
    void print() const;
    
    // Call backs that will be called by art.
    void preModuleConstruction (art::ModuleDescription const& md);
    void postModuleConstruction(art::ModuleDescription const& md);
    void preModuleBeginRun     (art::ModuleDescription const& md);
    void postModuleBeginRun    (art::ModuleDescription const& md);
    void postEndJob();
    
  private:
    using SeedMaster_t = SeedMaster<seed_t>;
    using EngineId = SeedMaster_t::EngineId;
    
    
    /// Class managing the seeds
    SeedMaster_t seeds;
    
    /// Helper to track state of art. It is legal to call getSeed only in certain states.
    SeedServiceHelper::ArtState state;
    
    /// Control the level of information messages.
    int verbosity = 0;
    bool bPrintEndOfJobSummary = false; ///< print a summary at the end of job
    
    // Main logic for computing and validating a seed.
    seed_t getSeed(EngineId const&);
    
    // Helper functions for all policies
    void ensureValidState() const;
    
  }; // class SeedService
  
  
  
  //================================================================
  template<class Stream>
  inline void SeedService::print(Stream& log) const {
    seeds.print(log);
  } // print(Stream)
  
} // namespace artext

DECLARE_ART_SERVICE(artext::SeedService, LEGACY)

#endif /* SeedService_SeedService_hh */
